#!/usr/bin/env python
# coding: utf-8

# In[1]:


#changement de répertoire
import os
os.chdir("D:/_Travaux/university/Cours_Universite/Supports_de_cours/DataAnalysis/RegressionLogistique/TD_9")

#chargement des données d'apprentissage
import pandas
DTrain = pandas.read_excel("infidelites_python.xlsx", sheet_name = "train")

#info
print(DTrain.info())

# In[2]:


#y (Infidelite) est la dernière colonne
yTrain = DTrain.iloc[:,-1]

#X (les autres) sont les variables qui précèdent la dernière
XTrain = DTrain.iloc[:,:-1]

#comptage des modalités de y
print(yTrain.value_counts())



# In[3]:

#chargement des données test
DTest = pandas.read_excel("infidelites_python.xlsx", sheet_name = "test")

#info
print(DTest.info())

# In[4]:


#y (Infidelite) est la dernière colonne
yTest = DTest.iloc[:,-1]

#X (les autres) sont les variables qui précèdent la dernière
XTest = DTest.iloc[:,:-1]

#comptage des modalités de y
print(yTest.value_counts())


## Régression logistique avec "statsmodels"


# In[5]:


#importation de la librairie de calcul
import statsmodels as sm

#vérification de version
print(sm.__version__)


# In[6]:


#importation de l'outil
from statsmodels.tools import add_constant

#données X avec la constante
XTrainBis = sm.tools.add_constant(XTrain)

#vérifier la structure
print(XTrainBis.info())


# In[7]:

#visualisation des premières lignes de la structure
#premières lignes
print(XTrainBis.head())


# ### Modélisation Logit
# Nous pouvons lancer la régression maintenant.

# In[8]:


#importation de la classe de calcul
from statsmodels.api import Logit

#régression logistique - on passe la cible et les explicatives
lr = Logit(endog=yTrain,exog=XTrainBis)

#lancer les calculs
#algorithme de Newton-Raphson utilisé par défaut
#https://www.statsmodels.org/stable/generated/statsmodels.discrete.discrete_model.Logit.fit.html
res = lr.fit()

#propriétés de l'objet résultat
#https://www.statsmodels.org/stable/generated/statsmodels.discrete.discrete_model.LogitResults.html
print(dir(res))


# In[9]:


#résumé des résultats
print(res.summary())


# On peut calculer les intervalles de confiance des coefficients pour un autre niveau de confiance, par ex. à 90%.

# In[10]:


#intervalle de confiance des coefficients à 90%
print(res.conf_int(alpha=0.1))


# ### Matrice de confusion en resubstitution
# Confronter les valeurs observées et prédites de la variable cible.

# In[11]:


#valeurs estimées par la régression en resubstitution
print(res.fittedvalues)


# Il s'agit en réalité des valeurs du LOGIT estimé. On peut le reproduire à partir des coefficients de la régression.

# In[12]:


#Voici les coefficients estimés
print(res.params)


# In[13]:


#description du premier individu
print(XTrainBis.iloc[0,:])


# In[14]:


#et si on fait le produit scalaire
import numpy
print(numpy.sum(res.params*XTrainBis.iloc[0,:]))


# In[15]:


#la règle d'affectation consiste à confronter le LOGIT à la valeur seuil 0
predResub = numpy.where(res.fittedvalues > 0, 1, 0)
print(predResub)


# In[16]:


#on peut en déduire la matrice de confusion
print(pandas.crosstab(yTrain,predResub))


# Laquelle peut être obtenue directement via une fonction de l'outil.

# In[17]:


#matrice de confusion en resubstitution directement fournie par l'outil
print(res.pred_table())


# ## Pseudo-R2
# Calcul des pseudo-R2, basés sur la confrontation entre la log-vraisemblance du modèle et de celle du modèle trivial composé uniquement de la constante (null model).

# In[18]:


#accès à la log-vraisemblance du modèle
print("Log-vraisemblance du modèle : %.4f" % (res.llf))

#log-vraisemblance du null modèle
print("Log-vraisemblance du null modèle : %.4f" % (res.llnull))


# In[19]:


#R2 de McFadden
R2MF = 1 - res.llf / res.llnull
print(R2MF)


# In[20]:


#qui est founi directement par l'outil
print(res.prsquared)


# Calcul du R2 de Cox et Snell

# In[21]:


#exponenielle de LL_null
L0 = numpy.exp(res.llnull)

#exponentielle de LL_modèle
La = numpy.exp(res.llf)

#taille de l'échantillon
n = DTrain.shape[0]

#R2 de Cox et Snell
R2CS = 1.0 - (L0 / La)**(2.0/n)
print("R2 de Cox - Snell : %.4f" % (R2CS))


# Calcul du R2 de Nagelkerke

# In[22]:


#max du R2 de COx-Snell
maxR2CS = 1.0 - (L0)**(2.0/n)

#R2 de Nagelkerke
R2N = R2CS / maxR2CS
print("R2 de Nagelkerke : %.4f" % (R2N))


# ## Evaluation basés sur les scores
# On se base sur les scores - c.-à-d. probabilité d'affectation à la modalité cible - pour évaluer la qualité d'approximation du modèle. Il faut déjà les calculer.

# In[23]:


#scores fournis par la régression
scores = lr.cdf(res.fittedvalues)
print(scores[:10])


# In[24]:


#vérifions le premières valeurs
s0 = 1.0/(1.0 + numpy.exp(-1.0 * res.fittedvalues[0]))
print("Score du 1er individu %.4f" % (s0))


# ### Diagramme de fiabilité
# 

# In[25]:


#data frame temporaire avec y et les scores
df = pandas.DataFrame({"y":yTrain,"score":scores})

#5 intervalles de largeur égales
intv = pandas.cut(df.score,bins=5,include_lowest=True)

#intégrées dans le df
df['intv'] = intv
print(df)


# In[26]:


#moyenne des scores par groupe
m_score = df.pivot_table(index="intv",values="score",aggfunc="mean")
print(m_score)


# In[27]:


#moyenne des y - qui équivaut à une proportion puisque 0/1
m_y = df.pivot_table(index="intv",values="y",aggfunc="mean")
print(m_y)


# In[28]:


#pouvoir faire apparaître le graphique dans le notebook
get_ipython().run_line_magic('matplotlib', 'inline')

#insérer le graphique
import matplotlib.pyplot as plt

#construire la diagonale
plt.plot(numpy.arange(0,1,0.1),numpy.arange(0,1,0.1),'b')

#rajouter notre diagramme
plt.plot(m_score,m_y,"go-")

#titre
plt.title("Diagramme de fiabilité")

#faire apparaître
plt.show()


# ### Test de Hosmer-Lemeshow

# In[29]:


#data frame temporaire avec y et les scores
df = pandas.DataFrame({"y":yTrain,"score":scores})

#10 intervalles de fréquences égales
intv = pandas.qcut(df.score,q=10)

#intégrées dans le df
df['intv'] = intv
print(df)


# In[30]:


#effectifs par groupe
n_tot = df.pivot_table(index="intv",values="y",aggfunc="count").values[:,0]
print(n_tot)


# In[31]:


#somme des scores par groupes
s_scores = df.pivot_table(index='intv',values="score",aggfunc="sum").values[:,0]
print(s_scores)


# In[32]:


#nombre de positifs par groupe
n_pos = df.pivot_table(index="intv",values="y",aggfunc="sum").values[:,0]
print(n_pos)


# In[33]:


#nombre de négatifs par groupe
n_neg = n_tot - n_pos
print(n_neg)


# In[34]:


C1 = numpy.sum((n_pos - s_scores)**2/s_scores)
print(C1)


# In[35]:


C2 = numpy.sum((n_neg - (n_tot - s_scores))**2/((n_tot - s_scores)))
print(C2)


# In[36]:


#statistique de Hosmer-Lemeshow
HL = C1 + C2
print(HL)


# In[37]:


#librairie scipy
import scipy

#probabilité critique
pvalue = 1.0 - scipy.stats.chi2.cdf(HL,8)
print(pvalue)


# ## Tests de significativité des coefficients

# ### Test de significativité globale de la régression
# Vérifier que la régression est globalement significative c.-à-d. tous les coefficients à l'exception de la constante sont tous nuls.

# #### Via le test du rapport de vraisemblance
# Basé sur la confrontation des déviances du modèle et du null modèle. On se rappelle que la déviance est égale à (-2) x log-vraisemblance.

# In[38]:


#déviance du modèle
dev_modele = (-2) * res.llf
print("Deviance du modèle %.4f : " % (dev_modele))

#déviance du modèle trivial
dev_null = (-2) * res.llnull
print("Deviance du modèle %.4f : " % (dev_null))

#statistique du rapport de vraisemblance
LR_stat = dev_null - dev_modele
print("Stat. du rapport de vraisemblance %.4f : " % (LR_stat))

#laquelle était fournie directement par l'objet
print("Stat. du rapport de vraisemblance via l'lobjet résultat : %.4f" % (res.llr))


# In[39]:


#degré de liberté du test (nb. de coef. estimé excepté la constante)
print(res.df_model)


# In[40]:


#p-value du test
pvalue = 1.0 - scipy.stats.chi2.cdf(res.llr,res.df_model)
print(pvalue)


# In[41]:


#laquelle était également fournie par l'objet
print(res.llr_pvalue)


# #### Via les comparaisons des critères AIC et BIC
# Une autre approche méconnue pour évaluer la qualité globale du modèle consiste à confronter les critères AIC ou BIC du modèle avec ceux du modèle trivial.

# In[42]:


#AIC du modèle
print("AIC du modèle : %.4f" % (res.aic))

#AIC du modèle trivial - 1 seul param. estimé, la constante
aic_null = (-2) * res.llnull + 2 * (1)
print("AIC du modèle trivial : %.4f" % (aic_null))


# Nous pouvons adopter la même démarche pour le critère BIC, qui nous indique que le modèle n'est pas si pertinent que cela.

# In[43]:


#BIC du modèle
print("BIC du modèle : %.4f" % (res.bic))

#BIC du modèle trivial - 1 seul param. estimé, la constante
bic_null = (-2) * res.llnull + numpy.log(n) * (1)
print("BIC du modèle trivial : %.4f" % (bic_null))


# ### Tester la significativité d'un coefficient

# Par exemple, pour évaluer le coefficient associé à la variable "Sex". Nous effectuons la régression en l'excluant.

# In[44]:


#data frame sans "Sex"
XTrain_wo_Sex = XTrainBis.drop(columns=['Sex'])
print(XTrain_wo_Sex.info())


# In[45]:


#régression sans Sex
lr_wo_Sex = Logit(yTrain,XTrain_wo_Sex)

#résultats
res_wo_Sex = lr_wo_Sex.fit()

#affichage
print(res_wo_Sex.summary())


# In[46]:


#statistique de test - différence entre les déviances
LR_Sex = (-2) * res_wo_Sex.llf - (-2) * res.llf
print(LR_Sex)


# In[47]:


#degré de liberté = 1 puisqu'un seul coef. retiré
ddl = res_wo_Sex.df_resid - res.df_resid
print("Degré de liberté du test : %.d" % (ddl))

#p-value
pvalue = 1.0 - scipy.stats.chi2.cdf(LR_Sex,ddl)
print("Probabilité critique : %.4f" % (pvalue))


# ### Tester un groupe de coefficients

# #### Le plus simple, passer par la vraisemblance

# In[48]:


#définir la matrice des X sans les 3 variables
XTrain_wo_3 = XTrainBis.drop(columns=['Sex','Education','Occupation'])
print(XTrain_wo_3.info())


# In[49]:


#régression sans les 3 variables
lr_wo_3 = Logit(yTrain,XTrain_wo_3)

#résultats
res_wo_3 = lr_wo_3.fit()

#statistique de test - différence entre les déviances
LR_3 = (-2) * res_wo_3.llf - (-2) * res.llf
print("Statistique de test : %.4f" % (LR_3))

#degré de liberté = 3 puisque 3 coef. retiré
ddl = res_wo_3.df_resid - res.df_resid
print("Degré de liberté du test : %.d" % (ddl))

#p-value
pvalue = 1.0 - scipy.stats.chi2.cdf(LR_3,ddl)
print("Probabilité critique : %.4f" % (pvalue))


# La nullité simultanée des coefficients de (Sex, Education, Occupation) n'est pas contredite par les données.

# ### Test de Wald -- Calcul matriciel

# In[50]:


#matrice de var-covar des coefs. estimés.
print(res.cov_params())

#récupération sous une forme matricielle
vcov = res.cov_params().values


# In[51]:


#nous avons bien le carré des ecarts-type estimés sur la diagionale
#à confronter avec les sorties de summary
print(numpy.sqrt(numpy.diagonal(vcov)))


# On va reproduire les calculs pour H0 : a_Sex = a_Education = a_Occupation = 0

# In[52]:


#indice des coefficients concernés
indices = [1,6,7]

#sous-matrice de var.covar
subset_vcov = numpy.zeros(shape=(3,3))
for i in range(3):
    for j in range(3):
        subset_vcov[i,j] = vcov[indices[i],indices[j]]

#vérification
print(subset_vcov)


# In[53]:


#inversion de cette matrice
inv_subset_vcov = numpy.linalg.inv(subset_vcov)
print(inv_subset_vcov)


# In[54]:


#coefficients estimés
a = res.params[indices].values

#produit matriciel
stat_3 = numpy.dot(a,numpy.dot(inv_subset_vcov,a))
print("Stat. de test : %.4f" % (stat_3))

#p-value (ddl = 3 puisque 3 coef. à tester)
pvalue = 1.0 - scipy.stats.chi2.cdf(stat_3,3)
print("p-value : %.4f" % (pvalue))


# #### Test de Wald avec l'outil dédié

# In[55]:


#matrice des coefficients à tester
R = [[0,1,0,0,0,0,0,0,0],[0,0,0,0,0,0,1,0,0],[0,0,0,0,0,0,0,1,0]]

#calculer la stat. de test
stat_3bis = res.wald_test(R)
print(stat_3bis)


# ## Performances prédictives
# Utilisation de l'échantillon test pour évaluer le modèle.

# ### Prédiction et matrice de confusion

# In[56]:


#préparation de l'échantillon test
#par adjonction de la constante
XTest_Bis = add_constant(XTest)
print(XTest_Bis.info())


# In[57]:


#calcul de la prédiction sur l'échantillon test
predProbaSm = res.predict(XTest_Bis)

#à l'évidence nous avons les probabilités d'affectation
print(predProbaSm.describe())


# In[58]:


#convertir en prédiction brute
predSm = numpy.where(predProbaSm > 0.5, 1, 0)
print(numpy.unique(predSm,return_counts=True))


# In[59]:


#matrice de confusion
mcSm = pandas.crosstab(yTest,predSm)
print(mcSm)


# In[60]:


#transformer en matrice Numpy
mcSmNumpy = mcSm.values

#taux de reconnaissance
accSm = numpy.sum(numpy.diagonal(mcSmNumpy))/numpy.sum(mcSmNumpy)
print("Taux de reconnaissance : %.4f" % (accSm))

#taux d'erreur
errSm = 1.0 - accSm
print("Taux d'erreur' : %.4f" % (errSm))


# ### Courbe ROC en test
# Construire la courbe ROC et calculer la valeur de l'AUC sur l'échantillon test. On utilise "scikit-learn" pour aller au plus court.

# In[61]:


#importer la librairie
import sklearn.metrics as metrics

#colonnes pour les courbes ROC
fprSm, tprSm, _ = metrics.roc_curve(yTest,predProbaSm,pos_label=1)

#graphique
#construire la diagonale
plt.plot(numpy.arange(0,1.1,0.1),numpy.arange(0,1.1,0.1),'b')

#rajouter notre diagramme
plt.plot(fprSm,tprSm,"g")

#titre
plt.title("Courbe ROC")

#faire apparaître
plt.show()


# In[62]:

#valeur de l'AUC
aucSm = metrics.roc_auc_score(yTest,predProbaSm)
print("AUC : %.4f" % (aucSm))


# # Régression logistique avec "scikit-learn"

# ## Importation et vérification de version

# In[63]:


#importation
import sklearn

#version
print(sklearn.__version__)


# ## Régression logistique avec scikit-learn

# In[64]:


#importation de la classe de calcul
from sklearn.linear_model import LogisticRegression

#instanciation
lrSk = LogisticRegression(penalty='none')

#lancement des calculs
#pas nécessaire de rajouter la constante
lrSk.fit(XTrain,yTrain)

#affichage des coefficients
print(pandas.DataFrame({"var":XTrain.columns,"coef":lrSk.coef_[0]}))


# In[65]:


#la constante
print(lrSk.intercept_)


# ### Régression logistique - Var. standardisées
# On va centrer et réduire les variables.

# In[66]:


#importation de l'outil
from sklearn import preprocessing

#instanciation
stds = preprocessing.StandardScaler()

#transformation
ZTrain = stds.fit_transform(XTrain)
print(scipy.stats.describe(ZTrain,axis=0,ddof=0))


# In[67]:


#instanciation
lrSkStd = LogisticRegression(penalty='none')

#lancement des calculs -- pas nécessaire de rajouter la constante
lrSkStd.fit(ZTrain,yTrain)

#affichage des coefficients
print(pandas.DataFrame({"var":XTrain.columns,"coef":lrSkStd.coef_[0]}))


# Pas de message d'erreur ou de warning cette fois-ci.
# Mais les coefficients doivent corrigés par les écarts-type

# In[68]:


#correction des coefficients - dé-standardisation
#par les écarts-type utilisés lors de la standardisation des variables
coefUnstd = lrSkStd.coef_[0] / stds.scale_

#affichage des coefficients corrigés
print(pandas.DataFrame({"var":XTrain.columns,"coef":coefUnstd}))


# La constante doit être corrigée également, en tenant de la constante estimées et des... montrer les formules...

# In[69]:


#pour la constante, l'opération est plus complexe
interceptUnStd = lrSkStd.intercept_ + numpy.sum(lrSkStd.coef_[0]*(-stds.mean_/stds.scale_))
print(interceptUnStd)


# ### Calcul de la log-vraisemblance
# On ne l'a pas directement, nous devons la calculer explicitement.

# In[70]:


#probabilités d'affectation
proba01 = lrSkStd.predict_proba(ZTrain)

#premières valeurs
print(proba01[:5,:])


# In[71]:


#récupération de la colonne n°1
proba1 = proba01[:,1]

#description stat
print(scipy.stats.describe(proba1))


# In[72]:


#log-vraisemblance
log_likelihood = numpy.sum(yTrain*numpy.log(proba1)+(1.0-yTrain)*numpy.log(1.0-proba1))
print(log_likelihood)


# ## Performances prédictives
# On applique le modèle sur l'échantillon test.

# In[73]:


#transformation de l'échantillon test
ZTest = stds.transform(XTest)

#stat. descriptives
print(scipy.stats.describe(ZTest,axis=0,ddof=0))


# In[74]:


#appliquer la prédiction
predSk = lrSkStd.predict(ZTest)
print(numpy.unique(predSk,return_counts=True))


# In[75]:


#matrice de confusion
print(metrics.confusion_matrix(yTest,predSk))


# In[76]:


#rapport sur la qualité de prédiction
print(metrics.classification_report(yTest,predSk))
